/* connect two port forward utility

   Copyright (c) 2003 Heikki Orsila <heikki.orsila@iki.fi>

   See COPYING file for terms (MIT license).
*/

#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#include <errno.h>
#include <fcntl.h>
#include <signal.h>
#include <sys/poll.h>
#include <sys/socket.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <unistd.h>
#include <sys/time.h>
#include <sys/wait.h>
#include <netinet/in.h>

#include "net.h"
#include "pipe.h"

extern int errno;

static int hello_from_src = 0;


static void pf_handle_client(int src_fd, int dst_fd)
{
	struct pipe p1;
	struct pipe p2;

	memset(&p1, 0, sizeof(struct pipe));
	memset(&p2, 0, sizeof(struct pipe));

	p1.src_fd = src_fd;
	p1.dst_fd = dst_fd;
	p2.src_fd = dst_fd;
	p2.dst_fd = src_fd;

	pmr_loop(&p1, &p2, NULL, NULL, NULL, NULL);

	close(p1.src_fd);
	close(p1.dst_fd);
	close(p2.src_fd);
	close(p2.dst_fd);
}

static int read_line(char *buf, int fd, int maxlen)
{
	size_t bytes_read = 0;
	int ret;
	while (1) {
		if (bytes_read >= ((size_t) maxlen))
			return 0;
		ret = read(fd, &buf[bytes_read], 1);
		if (ret < 0) {
			if (errno == EAGAIN || errno == EINTR) {
				continue;
			}
			perror("connecttwo: could not read line");
			return 0;
		} else if (ret == 0) {
			fprintf(stderr,
				"connecttwo: unexpected eof when reading a line\n");
			buf[bytes_read] = 0;
			return 0;
		}
		if (buf[bytes_read] == '\n') {
			buf[bytes_read] = 0;
			break;
		}
		bytes_read += ret;
	}
	return 1;
}

static int connect_two(char *src_name, char *src_port, char *dst_name,
		       char *dst_port)
{
	int src_fd;
	int dst_fd;
	pid_t child;
	int sleep_time;

	char line[256];

	sleep_time = 1;

	while (1) {

		if ((src_fd = net_open(src_name, src_port, "tcp", 0)) < 0) {
			fprintf(stderr,
				"connecttwo: couldn't connect to src: %s:%s. Sleeping %d seconds before re-try.\n",
				src_name, src_port, sleep_time);
			sleep(sleep_time);
			sleep_time *= 2;
			if (sleep_time > 64) {
				sleep_time = 64;
			}
			continue;
		}

		if (hello_from_src) {
			if (!read_line(line, src_fd, sizeof(line))) {
				fprintf(stderr,
					"connecttwo: didn't get hello from src. Sleeping %d seconds before re-try.\n",
					sleep_time);
				sleep(sleep_time);
				sleep_time *= 2;
				if (sleep_time > 64) {
					sleep_time = 64;
				}
				continue;
			}
		}

		if ((dst_fd = net_open(dst_name, dst_port, "tcp", 0)) < 0) {
			fprintf(stderr,
				"connecttwo: couldn't connect to dst: %s:%s. Sleeping %d seconds before re-try.\n",
				dst_name, dst_port, sleep_time);
			sleep(sleep_time);
			sleep_time *= 2;
			if (sleep_time > 64) {
				sleep_time = 64;
			}
			continue;
		}
		sleep_time = 1;

		child = fork();
		if (!child) {
			child = fork();
			if (!child) {
				pf_handle_client(src_fd, dst_fd);
				fprintf(stderr,
					"connecttwo: client handled\n");
			} else if (child == -1) {
				perror
				    ("connecttwo: children can not reproduce");
			}
			abort();

		} else if (child == -1) {
			perror
			    ("connecttwo: can not have children. client will die.");
		} else {
			wait(0);
		}
		close(src_fd);
		close(dst_fd);
	}

	close(src_fd);
	close(dst_fd);
	fprintf(stderr,
		"connecttwo: this should not happen. main loop discontinued.\n");
	return 1;
}

static int get_host_port(char **address, char **port, char *in)
{
	char *sep;
	char *nstr;

	if (!(nstr = strdup(in)))
		return 0;

	if (!(sep = strrchr(nstr, (int) ':'))) {
		return 0;
	}

	*sep = 0;

	*address = strdup(nstr);
	*port = strdup(sep + 1);

	free(nstr);

	if (!(*address) || !(*port)) {
		free(*address);
		free(*port);
		return 0;
	}

	return 1;
}


int main(int argc, char **argv)
{
	int i;
	char *src_name = 0;
	char *src_port = 0;
	char *dst_name = 0;
	char *dst_port = 0;

	for (i = 1; i < argc;) {

		if (!strcmp("-s", argv[i])) {
			if ((i + 1) >= argc) {
				fprintf(stderr,
					"not enough args. give source host:port\n");
				return -1;
			}
			if (!get_host_port
			    (&src_name, &src_port, argv[i + 1])) {
				return -1;
			}
			i += 2;

		} else if (!strcmp("-d", argv[i])) {
			if ((i + 1) >= argc) {
				fprintf(stderr,
					"not enough args. give destination host:port\n");
				return -1;
			}
			if (!get_host_port
			    (&dst_name, &dst_port, argv[i + 1])) {
				return -1;
			}
			i += 2;

		} else if (!strcmp("--hello-from-src", argv[i])) {
			hello_from_src = 1;
			i++;

		} else {
			fprintf(stderr, "illegal args (%s)\n", argv[i]);
			return -1;
		}
	}

	if (!src_port) {
		fprintf(stderr,
			"you must provide source with -s host:port!\n");
		return -1;
	}
	if (!dst_port) {
		fprintf(stderr,
			"you must provide destination with -d host:port!\n");
		return -1;
	}

	return connect_two(src_name, src_port, dst_name,
			   dst_port) ? -1 : 0;
}
