/* Copyright 2003 Heikki Orsila <heikki.orsila@iki.fi>

   See COPYING file for terms (MIT license).

   This module has been influenced by W. Richard Stevens networking code:
   W. Richard Stevens, UNIX Network Programming, ISBN 013490012X,
   Prentice Hall PTR.
*/

#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#include <sys/types.h>
#include <sys/socket.h>
#include <netdb.h>
#include <unistd.h>

#include <netinet/in.h>

#include "net.h"

int net_reuseaddr(int fd)
{
#ifdef SO_REUSEADDR
	int ret;
	int val = -1;
	ret = setsockopt(fd, SOL_SOCKET, SO_REUSEADDR, &val, sizeof(int));
	if (ret < 0) {
		perror("can't do SO_REUSEADDR");
	}
	return ret >= 0;
#else
	return 0;
#endif
}

static int get_sock_port(int fd, int addrlen)
{
	struct sockaddr_in *sockname;
	socklen_t socklen;
	int port;
	if (!(sockname = malloc(addrlen)))
		return 0;
	if (getsockname(fd, (struct sockaddr *) sockname, &socklen)) {
		free(sockname);
		return 0;
	}
	port = (int) ntohs(sockname->sin_port);
	free(sockname);
	return port;
}

int net_bind(int *port_result, char *hostname, char *port, char *protocol,
	     int reuse)
{
	int ret;
	int listenfd;
	struct addrinfo *res = 0;
	struct addrinfo *ressave;
	struct addrinfo hints;
	struct protoent *pe;
	char default_host[1024];

	if (!(pe = getprotobyname(protocol))) {
		perror("net_bind: can't get protocol number");
		return -1;
	}

	memset(&hints, 0, sizeof(hints));
	hints.ai_flags = AI_PASSIVE;
	hints.ai_family = AF_UNSPEC;
	hints.ai_socktype = SOCK_STREAM;
	hints.ai_protocol = pe->p_proto;

	if (!hostname) {
		char *host = getenv("INADDR_DEFAULT");
		if (host && strlen(host) > 0) {
			strncpy(default_host, host, sizeof(default_host));
			default_host[sizeof(default_host) - 1] = 0;
			hostname = default_host;
		}
	}

	ret = getaddrinfo(hostname, port, &hints, &res);
	if (ret) {
		fprintf(stderr, "getaddrinfo: %s (%s:%s)\n",
			gai_strerror(ret), hostname, port);
		return -1;
	}
	ressave = res;

	listenfd = -1;
	do {
		listenfd =
		    socket(res->ai_family, res->ai_socktype,
			   res->ai_protocol);
		if (listenfd < 0)
			continue;	/* error, try next one */

		if (reuse) {
			net_reuseaddr(listenfd);
		}

		if (bind(listenfd, res->ai_addr, res->ai_addrlen) == 0) {
			/* bind is successful! */
			if (port_result) {
				if (!
				    (*port_result =
				     get_sock_port(listenfd,
						   res->ai_addrlen))) {
					goto port_err;
				}
			}
			break;
		}
	      port_err:
		close(listenfd);
	} while ((res = res->ai_next) != 0);

	if (!res)
		listenfd = -1;

	freeaddrinfo(ressave);

	return listenfd;
}

int net_open(char *hostname, char *port, char *protocol, char *local_host)
{
	int ret;
	int sockfd;
	struct addrinfo *res = 0;
	struct addrinfo *ressave;
	struct addrinfo hints;
	struct protoent *pe;
	char default_host[1024];

	if (!(pe = getprotobyname(protocol))) {
		perror("net_open: can't get protocol number");
		return -1;
	}

	if (!local_host) {
		char *host = getenv("INADDR_DEFAULT");
		if (host && strlen(host) > 0) {
			strncpy(default_host, host, sizeof(default_host));
			default_host[sizeof(default_host) - 1] = 0;
			local_host = default_host;
		}
	}

	memset(&hints, 0, sizeof(hints));
	hints.ai_flags = 0;
	hints.ai_family = AF_UNSPEC;
	hints.ai_socktype = SOCK_STREAM;
	hints.ai_protocol = pe->p_proto;

	ret = getaddrinfo(hostname, port, &hints, &res);
	if (ret) {
		fprintf(stderr, "getaddrinfo: %s (%s:%s)\n",
			gai_strerror(ret), hostname, port);
		return -1;
	}
	ressave = res;

	sockfd = -1;
	do {
		sockfd =
		    socket(res->ai_family, res->ai_socktype,
			   res->ai_protocol);
		if (sockfd < 0)
			continue;

		if (local_host) {
			struct addrinfo ahints, *ares;
			memset(&ahints, 0, sizeof(struct addrinfo));
			ahints.ai_family = res->ai_family;
			ahints.ai_socktype = res->ai_socktype;
			ahints.ai_protocol = res->ai_protocol;
			ahints.ai_flags = AI_PASSIVE;
			if ((ret =
			     getaddrinfo(local_host, 0, &ahints, &ares))) {
				fprintf(stderr,
					"getaddrinfo on local host: %s (%s)\n",
					gai_strerror(ret), local_host);
				res = 0;
				break;
			}

			if (bind(sockfd, (struct sockaddr *) ares->ai_addr,
				 ares->ai_addrlen) < 0) {
				perror("bind failed to local host");
				freeaddrinfo(ares);
				continue;
			}
			freeaddrinfo(ares);
		}

		if (connect(sockfd, res->ai_addr, res->ai_addrlen) == 0)
			break;
		/* Can connect return EINTR? If so, we have a bug here. Anyway,
		   use of SIGALARM should be made possible here! Fix it. */
		close(sockfd);
	} while ((res = res->ai_next) != 0);

	if (!res) {
		sockfd = -1;
	}
	freeaddrinfo(ressave);
	return sockfd;
}

int net_info(struct net_host_id *id, char *hostname)
{
	int ret;
	struct addrinfo *res = 0;
	struct addrinfo hints;
	struct sockaddr_in *sockname;

	memset(&hints, 0, sizeof(hints));
	hints.ai_flags = 0;
	hints.ai_family = AF_UNSPEC;
	hints.ai_socktype = SOCK_STREAM;

	ret = getaddrinfo(hostname, 0, &hints, &res);
	if (ret) {
		fprintf(stderr,
			"net_info: getaddrinfo: %s (from host %s)\n",
			gai_strerror(ret), hostname);
		return 0;
	}
	id->len = 0;
	id->port = 0;
	sockname = (struct sockaddr_in *) res->ai_addr;
	switch (sockname->sin_family) {
	case AF_INET:
		id->len = 4;
		memcpy(id->addr, &sockname->sin_addr, id->len);
		ret = 1;
		break;
	case AF_INET6:
		id->len = 16;
		memcpy(id->addr, &sockname->sin_addr, id->len);
		ret = 1;
		break;
	default:
		ret = 0;
	}
	freeaddrinfo(res);
	return ret;
}

int net_are_same_ips(struct net_host_id *id1, struct net_host_id *id2)
{
	if (id1->len != id2->len)
		return 0;
	return memcmp(id1->addr, id2->addr, id1->len) == 0;
}

int net_print_id(char *str, struct net_host_id *id, int maxlen)
{
	char buf[64];
	int i;

	if (maxlen <= 0)
		return 0;
	str[0] = 0;
	/* 22 bytes reserved for a huger port number ;) */
	if (maxlen < (id->len * 4 + 1 + 22 + 1))
		return 0;
	/* inefficient. fix. */
	for (i = 0; i < (id->len - 1); i++) {
		sprintf(buf, "%d.", id->addr[i]);
		strcat(str, buf);
	}
	sprintf(buf, "%d:%d", id->addr[id->len - 1], id->port);
	strcat(str, buf);
	return 1;
}

int net_get_id(struct net_host_id *id, struct sockaddr *sockname)
{
	struct sockaddr_in *in = (struct sockaddr_in *) sockname;
	id->port = ntohs(in->sin_port);
	switch (in->sin_family) {
	case AF_INET:
		id->len = 4;
		memcpy(id->addr, &in->sin_addr, id->len);
		break;
	case AF_INET6:
		id->len = 16;
		memcpy(id->addr, &in->sin_addr, id->len);
		break;
	default:
		id->len = 0;
		id->port = 0;
		return 0;
	}
	return 1;
}
