#define _XOPEN_SOURCE 600

#include <stdio.h>
#include <stdlib.h>
#include <sys/time.h>
#include <sys/poll.h>
#include <errno.h>
#include <unistd.h>
#include <string.h>
#include <signal.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>

#include "pipe.h"
#include "accounting.h"
#include "unixatomic.h"


/* configuration variables */
static int default_interval = 2000;
int kill_on_eof = 1;
int pmr_max_rate = -1;
int pmr_shut_up = 0;
int pmr_accounting_interval = 512;
int pmr_print_pid = 1;

static int rate_read_bytes = 0;

struct timeval rate_time;


static void pmr_sigpipe(int sig)
{
	sig = sig;
	fprintf(stderr, "%d: SIGPIPE caught. Killing client.\n", getpid());
}


static int timetest(char *s, struct pipe *p, int force)
{
	struct timeval nt;
	int t;
	strcpy(s, "bandwidth: NaN");
	if (gettimeofday(&nt, 0)) {
		/* time failed. bandwidth = NaN. return false. */
		return 0;
	}
	t = 1000 * (nt.tv_sec - p->ot.tv_sec) + ((int) nt.tv_usec) / 1000 -
	    ((int) p->ot.tv_usec) / 1000;
	if (t < 0) {
		fprintf(stderr,
			"chronoton particles detected. clock ran backwards. k3wl!\n");
		t = default_interval + 1;
	}
	if (t > default_interval || force) {
		int order;
		double bw;
		char id[256];
		if (t) {
			bw = ((double) p->wbytes) * 1000.0f / ((double) t);
			order = 0;
			while (bw >= 1024.0f) {
				bw /= 1024.0f;
				order++;
				if (order == 4)
					break;
			}
			switch (order) {
			case 0:
				strcpy(id, "B");
				break;
			case 1:
				strcpy(id, "KiB");
				break;
			case 2:
				strcpy(id, "MiB");
				break;
			case 3:
				strcpy(id, "GiB");
				break;
			case 4:
				strcpy(id, "TiB");
				break;
			default:
				fprintf(stderr,
					"a bug in number order!\n");
				strcpy(id, "Strange Unit");
				break;
			}
			sprintf(s, "bandwidth: %.2f %s/s", bw, id);
		}
		p->ot = nt;
		p->wbytes = 0;
		return 1;
	}
	return 0;
}


int io_no_rate_limit(int fd, char *buf, int size,
		     ssize_t(*io_func) (int, void *, size_t))
{
	int ret;
	while ((ret = io_func(fd, buf, size)) < 0) {
		if (errno == EINTR)
			continue;
		return -1;
	}
	return ret;
}


int io_rate_limit(int fd, char *buf, int size,
		  ssize_t(*io_func) (int, void *, size_t))
{
	int ret;
	int to_read;
	int t;
	struct timeval new_rate_time;

	if (pmr_max_rate == -1)
		return io_no_rate_limit(fd, buf, size, io_func);

	if (gettimeofday(&new_rate_time, 0)) {
		perror
		    ("pmr: gettimeofday failed. can not limit rate. going max speed.");
		pmr_max_rate = -1;
		return io_no_rate_limit(fd, buf, size, io_func);
	}

	/* forgive me. function variables, such as rate_read_bytes, talks about
	   reading, but it does writing too depending on 'io_func' */

	if (rate_read_bytes > pmr_max_rate) {
		fprintf(stderr,
			"fatal error: rate_read_bytes > max_rate!\n");
		exit(-1);
	}

	if (rate_read_bytes == pmr_max_rate) {
		t = 1000 * (new_rate_time.tv_sec - rate_time.tv_sec) +
		    ((int) new_rate_time.tv_usec) / 1000 -
		    ((int) rate_time.tv_usec) / 1000;
		if (t < 0) {
			fprintf(stderr,
				"pmr: chronoton particles detected. clock ran backwards. k3wl!\n");
			t = default_interval + 1;
		}
		if (t < 1000) {
			usleep(1000000 - 1000 * t);
			if (gettimeofday(&new_rate_time, 0)) {
				perror
				    ("pmr: gettimeofday failed. can not limit rate. going max speed.");
				pmr_max_rate = -1;
				return io_no_rate_limit(fd, buf, size,
							io_func);
			}
		}
		rate_time = new_rate_time;
		rate_read_bytes = 0;
	}

	to_read = pmr_max_rate - rate_read_bytes;
	if (to_read > size)
		to_read = size;

	while ((ret = io_func(fd, buf, to_read)) < 0) {
		if (errno != EINTR)
			return -1;
	}
	rate_read_bytes += ret;
	return ret;
}


static int read_one(struct pipe *p)
{
	int ret;
	if (p->bytes == 0) {
		ret = io_rate_limit(p->src_fd, p->buf, p->bufsize, read);
		if (ret > 0) {
			p->bytes = ret;
		} else if (ret == 0) {
			p->eof_read = 1;
		} else {
			if (errno != EINTR && errno != EAGAIN) {
				perror("read error");
				p->eof_read = 1;
			}
		}
	} else {
		ret =
		    write(p->dst_fd, &p->buf[p->offs], p->bytes - p->offs);
		if (ret > 0) {
			p->wbytes += ret;
			p->tbytes += ret;
			p->offs += ret;
			p->accbytes += ret;	/* for accounting only */
		} else if (ret == 0) {
			fprintf(stderr,
				"%d: interesting: write returned 0\n",
				getpid());
			p->eof_read = 1;
			p->bytes = 0;
			p->offs = 0;
			return 0;
		} else if (ret < 0) {
			if (errno != EINTR && errno != EAGAIN) {
				perror("write error");
				p->eof_read = 1;
				p->bytes = 0;
				p->offs = 0;
				return 0;
			}
		}
		if (p->offs == p->bytes)
			p->bytes = p->offs = 0;
	}

	return 1;
}


void pmr_loop(struct pipe *p1, struct pipe *p2, char *host1, char *file1,
	      char *host2, char *file2)
{
	const int aligned_size = 8192;
	char *aligned_buf;
	void *real_buf;
	struct sigaction act;
	struct timeval ot;
	struct timeval nt;
	int ret;
	char info[256];
	int valid_time = 1;
	int p1_active;
	int p2_active;

	struct pollfd pfds[4];
	int something_polled;
	int i;

	/* allocate _two_ pipe buffers */
	if (posix_memalign(&real_buf, 8192, 2 * aligned_size)) {
		fprintf(stderr, "not enough memory: %s\n",
			strerror(errno));
		exit(-1);
	}
	aligned_buf = real_buf;
	if (p1) {
		p1->buf = aligned_buf;
		p1->bufsize = aligned_size;
	}
	if (p2) {
		p2->buf = aligned_buf + aligned_size;
		p2->bufsize = aligned_size;
	}

	memset(&act, 0, sizeof(act));
	act.sa_handler = pmr_sigpipe;
	sigaction(SIGPIPE, &act, 0);

	if (gettimeofday(&ot, 0)) {
		memset(&ot, 0, sizeof(ot));
		perror("gettimeofday() failed");
		valid_time = 0;
	}
	if (p1) {
		p1->ot = ot;
		p1->vot = ot;
	}
	if (p2) {
		p2->ot = ot;
		p2->vot = ot;
	}

	p1_active = (p1 != NULL);
	p2_active = (p2 != NULL);

	while (p1_active || p2_active) {

		int npfds;

		if (p1 && p1->eof_read && p1->bytes == 0)
			p1_active = 0;

		if (p2 && p2->eof_read && p2->bytes == 0)
			p2_active = 0;

		if (kill_on_eof) {
			if (p2 == NULL) {
				if (!p1_active)
					break;
			} else {
				if ((!p1_active || !p2_active)
				    && p1->bytes == 0 && p2->bytes == 0)
					break;
			}
		}

		memset(pfds, 0, sizeof(pfds));

		something_polled = 0;

		npfds = 2 + (p2_active ? 2 : 0);
		if (p1->eof_read == 0 && p1->bytes == 0) {
			pfds[0].fd = p1->src_fd;
			pfds[0].events = POLLIN;
			something_polled = 1;
		}
		if (p1->bytes) {
			pfds[1].fd = p1->dst_fd;
			pfds[1].events = POLLOUT;
			something_polled = 1;
		}

		if (p2_active && p2->eof_read == 0 && p2->bytes == 0) {
			pfds[2].fd = p2->src_fd;
			pfds[2].events = POLLIN;
			something_polled = 1;
		}
		if (p2_active && p2->bytes) {
			pfds[3].fd = p2->dst_fd;
			pfds[3].events = POLLOUT;
			something_polled = 1;
		}

		if (!something_polled)
			break;

		ret = poll(pfds, npfds, -1);

		if (ret > 0) {
			if (p1_active) {
				for (i = 0; i < 2; i++) {
					if (pfds[i].
					    revents & (pfds[i].
						       events | POLLERR |
						       POLLHUP)) {
						if (!read_one(p1))
							p1_active = 0;
						break;
					}
				}
			}

			if (p2_active) {
				for (i = 2; i < 4; i++) {
					if (pfds[i].
					    revents & (pfds[i].
						       events | POLLERR |
						       POLLHUP)) {
						if (!read_one(p2))
							p2_active = 0;
						break;
					}
				}
			}

		} else if (ret < 0) {
			if (errno != EINTR) {
				perror("poll error");
				goto out;
			}
		}

		if (valid_time) {
			if (p1_active && timetest(info, p1, 0)) {
				char byte_info[256];
				sprintf(byte_info, "\tbytes: %lld",
					p1->tbytes);
				strcat(info, byte_info);
				if (pmr_shut_up == 0) {
					if (pmr_print_pid)
						fprintf(stderr,
							"%d ->: %s\n",
							getpid(), info);
					else
						fprintf(stderr, "->: %s\n",
							info);
				}
			}
			if (p2_active && timetest(info, p2, 0)) {
				char byte_info[256];
				sprintf(byte_info, "\tbytes: %lld",
					p2->tbytes);
				strcat(info, byte_info);
				if (pmr_shut_up == 0) {
					if (pmr_print_pid)
						fprintf(stderr,
							"%d <-: %s\n",
							getpid(), info);
					else
						fprintf(stderr, "<-: %s\n",
							info);
				}
			}
			if (file1 != NULL && gettimeofday(&nt, 0) == 0) {
				if ((nt.tv_sec - ot.tv_sec) >=
				    pmr_accounting_interval) {
					long long accbytes = 0;
					long long hbytes;
					if (p1) {
						accbytes += p1->accbytes;
						p1->accbytes = 0;
					}
					if (p2) {
						accbytes += p2->accbytes;
						p2->accbytes = 0;
					}

					if ((hbytes =
					     account_increment(host1,
							       file1,
							       accbytes)) <
					    0) {
						fprintf(stderr,
							"was unable to increment host %s\n",
							host1);
					} else {
						fprintf(stderr,
							"host %s incremented by %.1f mebibytes to %.1f mebibytes (%lld bytes)\n",
							host1,
							((double) accbytes)
							/ (1024.0 *
							   1024.0),
							((double) hbytes) /
							(1024.0 * 1024.0),
							hbytes);
					}
					if (host2) {
						if ((hbytes =
						     account_increment
						     (host2, file2,
						      accbytes)) < 0) {
							fprintf(stderr,
								"was unable to increment host %s\n",
								host2);
						} else {
							fprintf(stderr,
								"host %s incremented by %.1f mebibytes to %.1f mebibytes (%lld bytes)\n",
								host2,
								((double)
								 accbytes)
								/ (1024.0 *
								   1024.0),
								((double)
								 hbytes) /
								(1024.0 *
								 1024.0),
								hbytes);
						}
					}
					ot = nt;
				}
			}
		}
	}

      out:

	if (p1) {
		p1->wbytes = p1->tbytes;
		p1->ot = p1->vot;
		timetest(info, p1, 1);
		if (pmr_shut_up == 0) {
			fprintf(stderr, "%d ->: average %s\n", getpid(),
				info);
			fprintf(stderr, "%d ->: total bytes: %lld\n",
				getpid(), p1->tbytes);
		}
	}
	if (p2) {
		p2->wbytes = p2->tbytes;
		p2->ot = p2->vot;
		timetest(info, p2, 1);
		if (pmr_shut_up == 0) {
			fprintf(stderr, "%d <-: average %s\n", getpid(),
				info);
			fprintf(stderr, "%d <-: total bytes: %lld\n",
				getpid(), p2->tbytes);
		}
	}
	fflush(stderr);
	free(real_buf);
}
